/*
  The Next Generation Beacon software package main file - AD9912_FreqCarrier_Off.c

  Small utility program to load an AD9912 with a frequency and turn it off

    FTW = 2^48 * Freq/1000000000, i.e. 48 bits involved

Copyright
  Bo Hansen, OZ2M

Compiler: AVR Studio 7.0.1188, Windows 10

MCU: Atmel Atmega128A, 8 MHz

Configuration
- Analog Devices AD9912 DDS Port A 7 to 1
    PA1 DDS RESET, output
    PA2 DDS SDO, input
    PA3 DDS SDIO, output/[input]
    PA4 DDS SCLK, output
    PA5 DDS CSB, output
    PA6 DDS IO_UPDATE, output
    PA7 DDS PWRDOWN, output
- DIP8 switch attached to Port C
    PC0 to PC7, output
- NMEA GPS signal Port D
    PD2(RXD1), input
- Auxiliary I/O
    PD4 AUX IN1, input
    PD5 AUX IN2, input
    PD6 AUX OUT, output
    Port F, PF0 to PF7 on J12
- FTDI FT232R USB <-> 3,3 V
    PE0(RDX0), input
    PE1(TXD0), output
    PE2 serial detect, input

Homepage: www.rudius.net/oz2m/ngnb
*/

#define F_CPU 8000000

#define DDSPort PORTA
#define DDSReset 1
#define DDSSDO 2                        // Input, optional
#define DDSSDIO 3                       // "MOSI" to DDS slave
#define DDSSCLK 4
#define DDSCSB 5
#define DDSIO_UPDATE 6

#define AuxPort PORTF

#define SET(x,y) (x |=(1<<y))           //-Bit set/clear macros
#define CLR(x,y) (x &= (~(1<<y)))       // |
#define CHK(x,y) (x & (1<<y))           // |
#define TOG(x,y) (x^=(1<<y))            //-+

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
#include <avr/io.h>
#include <util/delay.h>
#include <avr/eeprom.h>

// Communication variables
char Buffer[30];
uint8_t BufferCount = 0;
char Ch;

unsigned char FTW[2][6] = { {0, 0, 0, 0, 0, 0} , {0x0C, 0xE6, 0xEE, 0xB7, 0x02, 0x60} };

// Help vars
int8_t I;

// ------------------- DDS routines ---------------------------------------------------------------------------------------
void DDSBitBangOut(unsigned char Value)                   // Bit bang way to load the DDS with the bits
{
    for (char I = 1; I <= 8; I++)                            // Setup a loop of 8 iterations, one for each bit
    {
        if (Value > 127)                                       // Test MSB
            SET(DDSPort, DDSSDIO);                               // If it is a 1 (i.e. 1XXXXXXX), set the bit bang MOSI high
        else
            CLR(DDSPort, DDSSDIO);                               // If it is a 1 (i.e. 1XXXXXXX), set the bit bang MOSI low

        SET(DDSPort, DDSSCLK);                                 // Set clock high

        Value = Value << 1;                                    // Shift MSB-1 to become the next MSB bit in TheByte

        CLR(DDSPort, DDSSCLK);                                 // Set clock low
    }
}

void DDSFreq(unsigned char Nbr)                            // Sends frequency tuning word FTW to DDS
{
    CLR(DDSPort, DDSCSB);                                    // Select "this" DDS

    DDSBitBangOut(0x61);                                     // Send freq. address word, 16 bits, DDS data sheet p. 28 table 9 and p. 29 table 10,
    DDSBitBangOut(0xAB);                                     // 011 0 bin + 01AB hex

    DDSBitBangOut(FTW[Nbr][0]);
    DDSBitBangOut(FTW[Nbr][1]);
    DDSBitBangOut(FTW[Nbr][2]);
    DDSBitBangOut(FTW[Nbr][3]);
    DDSBitBangOut(FTW[Nbr][4]);
    DDSBitBangOut(FTW[Nbr][5]);

    SET(DDSPort, DDSIO_UPDATE);                              // Activate "new" frequency
    _delay_us(1);                                            // small pause here needed??
    CLR(DDSPort, DDSIO_UPDATE);                              // Put "transfer data" back into idle
    SET(DDSPort, DDSCSB);                                    // Deselect "this" DDS
}


void DDSDACCurrent()
{
    CLR(DDSPort, DDSCSB);                                    // Select "this" DDS
    DDSBitBangOut(0x04);                                     // Address 0x040B
    DDSBitBangOut(0x0B);
    DDSBitBangOut(eeprom_read_byte((uint8_t*) 6));

    DDSBitBangOut(0x04);                                     // Address 0x040C
    DDSBitBangOut(0x0C);
    DDSBitBangOut(eeprom_read_byte((uint8_t*) 7));
}

// ------------------- Serial port routines ------------------------------------------------------------------------

int sendSerialChar(int Data)                                                             // See example in section 20.6.1
{
    while (!(UCSR0A & (1 << UDRE0)));                                                      // Wait for empty transmit buffer
    UDR0 = Data;                                                                           // Put data into buffer, sends the data

    return 0;
}

void Reconfig_printf()                                                                  // Reconfigure printf to use sendSerialChar that uses USART0
{
    static FILE uart_str = FDEV_SETUP_STREAM(sendSerialChar, NULL, _FDEV_SETUP_RW);
    stdout = stdin = &uart_str;
}

// ------------------- H/W routine -----------------------------------------------------------------------------------------
void HardwareInit()
{
    // Ports
    DDRA = 0b11111010;                                       // Config port A
    DDRC = 0b00000000;                                       // Config port C all as input
    PORTC = 0b11111111;                                      // Enable port C enable pull-ups
    DDRD = 0b11001011;                                       // Config port D
    DDRE = 0b11111010;                                       // Config port E
    DDRF = 0b11111111;                                       // Config port F all as outputs

    // DDS
    CLR(DDSPort, DDSReset);                                                              // Reset the DDS
    _delay_ms(10);
    SET(DDSPort, DDSReset);
    _delay_ms(10);
    CLR(DDSPort, DDSReset);
    _delay_ms(10);

    DDSBitBangOut(0x10);                                     // Init control word, 16 bits, see DDS data sheet p. 28 and p. 33 table 13, MSB first, SDIO used
    DDSBitBangOut(0x00);
    DDSFreq(0);                                              // Stop carrier while setting up

    // Test serial port UART0, 9600 baud, 8N1 can be used to echo stuff
    UCSR0B = 0b00000000;                                     // Section 20.9.3, disable interrupts, receiving and transmitting while setting up
    UCSR0A = 0b00000000;                                     // Section 20.9.2, U2X=0 i.e. disabled
    UCSR0C = 0b00000110;                                     // Section 20.9.4, 0, async UMSELn=0, parity N UPM=00, stop 1 USBS=0, bits 8 UCSZ=11, rising edge UCPOL=0
    UBRR0L = 51;                                             // Table 20-11, 9600 baud at 8 MHz clock, UX2=0
    UBRR0H = 0;                                              // High part of baud rate

    UCSR0B = 0b00011000;                                     // Serial management , section 20.9.3, enable RX Interrupt bit 7=1, RX enable bit4 RXEN0=1, TX enable TXEN0=1, UCSZ12=0
}

void PrintInfo()
{
    printf("\n\n\n");
    printf("AD9912 carrier test program v. 1.5.0\n");
    printf("www.rudius.net/oz2m/ngnb/dds_sw_freqcarrier_off.htm\n");
    printf("\n");
    printf("Syntax:\n");
    printf("  f FREQ\n");
    printf("      where FREQ is in Hz, FREQ <= 440000000\n\n");
    printf("  fd FREQ.DEC\n");
    printf("      where FREQ.DEC is in Hz and ALWAYS with six decimals, FREQ <= 440000000.999999\n\n");
    printf("  fs\n");
    printf("      to save the current FTW/frequency and use it when starting up\n\n");
    printf("  ftw FTW0 FTW1 FTW2 FTW3 FTW4 FTW5\n");
    printf("      to set the FTW where FTW# is hex bytes without 0x\n\n");
    printf("  rd ftw\n");
    printf("      to print the current FTW in hex bytes without 0x\n\n");
    printf("  wr INST VAL0 [VAL1 .. VAL5]\n");
    printf("      where INST is the instruction and VAL# is the register value. Both in hex\n\n");
    printf("  on\n");
    printf("      to turn the carrier on again at latest frequency\n\n");
    printf("  off\n");
    printf("      to turn the carrier off\n\n");
    printf("  dac CURRENT\n");
    printf("      to set the DAC current from 0 1023 to \n\n");
}

void CommandPaser(const char *string_to_parse)      // Command parser, processes command, return 1 if "done" management otherwise return 0
{
    static const char command_freq[] = "f ";        // Write frequency to DDS input is in decimal and Hz resolution
    static const char command_freqDec[] = "fd ";    // Write frequency with six decimals to DDS input is in Hz, e.g. 50399000.123456
    static const char command_ftwstart[] = "fs";    // Set FTW/command_ftwstart at startup
    static const char command_ftw[] = "ftw ";       // Write ftw to DDS input is 6 ftw bytes in hex

    static const char command_rdftw[] = "rd ftw";   // No carrier

    static const char command_wrreg[] = "wr ";      // Write value to register

    static const char command_on[] = "on";          // Carrier on
    static const char command_off[] = "off";        // No carrier

    static const char command_dac[] = "dac";        // DAC current

    const long int BitsClock = 281475;              // (2^48) / 1 GHz +1
    int8_t I;
    long int Freq, FDec;                            // Input frequency data value
    long long int FreqBits, FreqBitsDec;
    uint16_t ftw0, ftw1, ftw2, ftw3, ftw4, ftw5;
    uint8_t count;
    uint16_t head;
    uint16_t val0, val1, val2, val3, val4, val5;
    uint8_t val[6];
    uint16_t dacCurrent;

    if (0 == strncmp(command_freq, string_to_parse, sizeof(command_freq) - 1)) // Write command check
    {
        if (1 == sscanf(&string_to_parse[sizeof(command_freq) - 1], "%lu", &Freq))
        {
            if (Freq <= 440000000LU)
            {
                FreqBits = (long long int) Freq * BitsClock - Freq / 42 + Freq / 1922 - Freq / 8965706;

                for (I = 5; I >= 0; I--)
                {
                    FTW[1][I] = FreqBits;
                    FreqBits = FreqBits >> 8;
                }

                DDSFreq(1);
                printf("Freq: %lu Hz -> FTW: ", Freq);
                printf("%02X %02X %02X %02X %02X %02X\n", FTW[1][0], FTW[1][1], FTW[1][2], FTW[1][3], FTW[1][4], FTW[1][5]);
            }
            else
                printf("Invalid frequency\n");
        }
    }
    else

        if (0 == strncmp(command_freqDec, string_to_parse, sizeof(command_freqDec) - 1)) // Write command check
        {
            if (2 == sscanf(&string_to_parse[sizeof(command_freqDec) - 1], "%lu.%lu", &Freq, &FDec))
            {
                if (Freq <= 440000000LU)
                {
                    FreqBits = (long long int) Freq * BitsClock - Freq / 42 + Freq / 1922 - Freq / 8965706;
                    FreqBitsDec = (float) 0.2814749767 * FDec;

                    FreqBits += FreqBitsDec;

                    for (I = 5; I >= 0; I--)
                    {
                        FTW[1][I] = FreqBits;
                        FreqBits = FreqBits >> 8;
                    }

                    DDSFreq(1);
                    printf("Freq: %lu.%lu Hz -> FTW: ", Freq, FDec);
                    printf("%02X %02X %02X %02X %02X %02X\n", FTW[1][0], FTW[1][1], FTW[1][2], FTW[1][3], FTW[1][4], FTW[1][5]);
                }
                else
                    printf("Invalid frequency\n");
            }
        }
        else

            if (0 == strncmp(command_ftwstart, string_to_parse, sizeof(command_ftwstart) - 1)) // fs command check
            {
                for (I = 5; I >= 0; I--)
                    eeprom_write_byte((uint8_t*) (uint16_t) I, FTW[1][I]);
                printf("OK\n");
            }
            else

                if (0 == strncmp(command_ftw, string_to_parse, sizeof(command_ftw) - 1)) // ftw command check
                {
                    if (6 == sscanf(&string_to_parse[sizeof(command_ftw) - 1], "%x %x %x %x %x %x", &ftw0, &ftw1, &ftw2, &ftw3, &ftw4, &ftw5))
                    {
                        FTW[1][0] = ftw0;
                        FTW[1][1] = ftw1;
                        FTW[1][2] = ftw2;
                        FTW[1][3] = ftw3;
                        FTW[1][4] = ftw4;
                        FTW[1][5] = ftw5;

                        DDSFreq(1);
                        printf("FTW: %02X %02X %02X %02X %02X %02X\n", FTW[1][0], FTW[1][1], FTW[1][2], FTW[1][3], FTW[1][4], FTW[1][5]);
                    }
                    else
                        printf("Invalid frequency\n");
                }
                else

                    if (0 == strncmp(command_wrreg, string_to_parse, sizeof(command_wrreg) - 1))   // Write register command check
                    {
                        count = sscanf(&string_to_parse[sizeof(command_wrreg) - 1], "%x %x %x %x %x %x %x", &head, &val0, &val1, &val2, &val3, &val4, &val5);
                        if (count > 1)
                        {
                            CLR(DDSPort, DDSCSB);                                    // Select "this" DDS
                            DDSBitBangOut((head >> 8) & 0xFF);                       // Send instruction header word, 16 bits, DDS data sheet p. 28 table 9 and p. 29 table 10,
                            DDSBitBangOut(head & 0xFF);

                            val[0] = val0;
                            val[1] = val1;
                            val[2] = val2;
                            val[3] = val3;
                            val[4] = val4;
                            val[5] = val5;

                            count--;                                                 // Remove instruction header count
                            for (I = 0; I < count; I++)
                                DDSBitBangOut(val[I]);

                            SET(DDSPort, DDSIO_UPDATE);                              // Activate "new" input
                            _delay_us(1);                                            // small pause here needed??
                            CLR(DDSPort, DDSIO_UPDATE);                              // Put "transfer data" back into idle
                            SET(DDSPort, DDSCSB);                                    // Deselect "this" DDS

                            printf("OK\n");
                        }
                        else
                            printf("Invalid inst value(s)\n");
                    }
                    else

                        if (0 == strncmp(command_rdftw, string_to_parse, sizeof(command_rdftw) - 1)) // "rd ftw" command check
                        {
                            printf("FTW: %02X %02X %02X %02X %02X %02X\n", FTW[1][0], FTW[1][1], FTW[1][2], FTW[1][3], FTW[1][4], FTW[1][5]);
                            printf("OK\n");
                        }
                        else

                            if (0 == strncmp(command_on, string_to_parse, sizeof(command_on) - 1)) // "on" command check
                            {
                                DDSFreq(1);                                                           // Switch carrier of
                                printf("OK\n");
                            }
                            else

                                if (0 == strncmp(command_off, string_to_parse, sizeof(command_off) - 1)) // "off" command check
                                {
                                    DDSFreq(0);                                                           // Switch carrier of
                                    printf("OK\n");
                                }
                                else

                                    if (0 == strncmp(command_dac, string_to_parse, sizeof(command_dac) - 1)) // DAC current command check
                                    {
                                        if (1 == sscanf(&string_to_parse[sizeof(command_dac) - 1], "%d", &dacCurrent))
                                        {
                                            if (dacCurrent <= 0x3FF)
                                            {
                                                for (I = 0; I < 2; I++)
                                                    eeprom_write_byte((uint8_t*) (uint16_t) I + 6, (dacCurrent >> (I * 8)) & 0xFF);

                                                DDSDACCurrent();

                                                printf("OK\n");
                                            }
                                            else
                                                printf("Invalid DAC current value\n");
                                        }
                                    }
                                    else
                                        PrintInfo();                                                          // Unknown info so print info text
}

int main()
{
    Reconfig_printf();                                       // printf is redirected to send output to USART0, i.e. management port

    HardwareInit();

    PrintInfo();

    // Load startup frequency
    for (I = 0; I < 6; I++)
        FTW[1][I] = eeprom_read_byte((uint8_t*) (uint16_t) I);
    DDSFreq(1);

    DDSDACCurrent();

    SET(DDSPort, DDSIO_UPDATE);                   // Activate "new" frequency
    _delay_us(1);                                 // small pause here needed??
    CLR(DDSPort, DDSIO_UPDATE);                   // Put "transfer data" back into idle
    SET(DDSPort, DDSCSB);                         // Deselect "this" DDS

    while (1 == 1)
    {
        if (CHK(UCSR0A, RXC0))                              // Check if character received
        {
            Ch = UDR0;                                          // Get received char

            if (Ch != '\n')
            {
                // Collect incoming chars until LF
                Buffer[BufferCount] = Ch;
                BufferCount++;
                Buffer[BufferCount] = 0;                         // Set EOS sz
            }
            else
            {
                CommandPaser(Buffer);
                BufferCount = 0;                                 // Reset Buffer counter
                Buffer[BufferCount] = 0;                         // Set EOS sz
            }
        }
    }
}